@@ -1,4 +1,4 @@
-#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -28,14 +28,15 @@ expression only stuff.
 from abc import abstractmethod
 
 from nuitka import Options, Tracing, TreeXML, Variables
-from nuitka.__past__ import (  # pylint: disable=I0021,redefined-builtin
-    intern,
-    iterItems
-)
-from nuitka.Errors import NuitkaNodeError
+from nuitka.__past__ import iterItems
+from nuitka.Errors import NuitkaNodeDesignError, NuitkaNodeError
 from nuitka.PythonVersions import python_version
 from nuitka.SourceCodeReferences import SourceCodeReference
-from nuitka.utils.InstanceCounters import counted_del, counted_init
+from nuitka.utils.InstanceCounters import (
+    counted_del,
+    counted_init,
+    isCountingInstances,
+)
 
 from .FutureSpecs import fromFlags
 from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions
@@ -46,7 +47,7 @@ class NodeBase(NodeMetaClassBase):
     __slots__ = "parent", "source_ref"
 
     # Avoid the attribute unless it's really necessary.
-    if Options.isFullCompat():
+    if Options.is_fullcompat:
         __slots__ += ("effective_source_ref",)
 
     # String to identify the node class, to be consistent with its name.
@@ -65,7 +66,8 @@ class NodeBase(NodeMetaClassBase):
 
         self.source_ref = source_ref
 
-    __del__ = counted_del()
+    if isCountingInstances():
+        __del__ = counted_del()
 
     @abstractmethod
     def finalize(self):
@@ -75,53 +77,42 @@ class NodeBase(NodeMetaClassBase):
         return "<Node %s>" % (self.getDescription())
 
     def getDescription(self):
-        """ Description of the node, intended for use in __repr__ and
-            graphical display.
+        """Description of the node, intended for use in __repr__ and
+        graphical display.
 
         """
-        details = self.getDetails()
+        details = self.getDetailsForDisplay()
 
         if details:
-            return "'%s' with %s" % (self.kind, self.getDetails())
+            return "'%s' with %s" % (self.kind, details)
         else:
             return "'%s'" % self.kind
 
     def getDetails(self):
-        """ Details of the node, intended for re-creation.
+        """Details of the node, intended for re-creation.
 
-            We are not using the pickle mechanisms, but this is basically
-            part of what the constructor call needs. Real children will
-            also be added.
+        We are not using the pickle mechanisms, but this is basically
+        part of what the constructor call needs. Real children will
+        also be added.
 
         """
         # Virtual method, pylint: disable=no-self-use
         return {}
 
     def getDetailsForDisplay(self):
-        """ Details of the node, intended for use in __repr__ and dumps.
+        """Details of the node, intended for use in __repr__ and dumps.
 
-            This is also used for XML.
+        This is also used for XML.
         """
         return self.getDetails()
 
-
-    def getDetail(self):
-        """ Details of the node, intended for use in __repr__ and graphical
-            display.
-
-        """
-        return str(self.getDetails())[1:-1]
-
     def getCloneArgs(self):
         return self.getDetails()
 
     def makeClone(self):
         try:
             # Using star dictionary arguments here for generic use.
-            result = self.__class__(
-                source_ref = self.source_ref,
-                **self.getCloneArgs()
-            )
+            result = self.__class__(source_ref=self.source_ref, **self.getCloneArgs())
         except TypeError as e:
             raise NuitkaNodeError("Problem cloning node", self, e)
 
@@ -138,10 +129,7 @@ class NodeBase(NodeMetaClassBase):
 
         try:
             # Using star dictionary arguments here for generic use.
-            result = self.__class__(
-                source_ref = self.source_ref,
-                **args
-            )
+            result = self.__class__(source_ref=self.source_ref, **args)
         except TypeError as e:
             raise NuitkaNodeError("Problem cloning node", self, e)
 
@@ -153,20 +141,16 @@ class NodeBase(NodeMetaClassBase):
         return result
 
     def getParent(self):
-        """ Parent of the node. Every node except modules have to have a parent.
-
-        """
+        """Parent of the node. Every node except modules have to have a parent."""
 
         if self.parent is None and not self.isCompiledPythonModule():
             # print self.getVisitableNodesNamed()
-            assert False, (self,  self.source_ref)
+            assert False, (self, self.source_ref)
 
         return self.parent
 
     def getChildName(self):
-        """ Return the role in the current parent, subject to changes.
-
-        """
+        """Return the role in the current parent, subject to changes."""
         parent = self.getParent()
 
         for key, value in parent.getVisitableNodesNamed():
@@ -183,29 +167,26 @@ class NodeBase(NodeMetaClassBase):
         child_name = self.getChildName()
 
         if hasattr(self.parent, "nice_children"):
-            return self.parent.nice_children[self.parent.named_children.index(child_name)]
+            return self.parent.nice_children[
+                self.parent.named_children.index(child_name)
+            ]
         elif hasattr(self.parent, "nice_child"):
             return self.parent.nice_child
         else:
             return child_name
 
     def getParentFunction(self):
-        """ Return the parent that is a function.
-
-        """
+        """Return the parent that is a function."""
 
         parent = self.getParent()
 
-        while parent is not None and \
-              not parent.isExpressionFunctionBodyBase():
+        while parent is not None and not parent.isExpressionFunctionBodyBase():
             parent = parent.getParent()
 
         return parent
 
     def getParentModule(self):
-        """ Return the parent that is module.
-
-        """
+        """Return the parent that is module."""
         parent = self
 
         while not parent.isCompiledPythonModule():
@@ -233,8 +214,10 @@ class NodeBase(NodeMetaClassBase):
     def getParentReturnConsumer(self):
         parent = self.getParent()
 
-        while not parent.isParentVariableProvider() and \
-              not parent.isExpressionOutlineBody():
+        while (
+            not parent.isParentVariableProvider()
+            and not parent.isExpressionOutlineBody()
+        ):
             parent = parent.getParent()
 
         return parent
@@ -258,42 +241,40 @@ class NodeBase(NodeMetaClassBase):
         return self.source_ref
 
     def setCompatibleSourceReference(self, source_ref):
-        """ Bug compatible line numbers information.
+        """Bug compatible line numbers information.
 
-            As CPython outputs the last bit of bytecode executed, and not the
-            line of the operation. For example calls, output the line of the
-            last argument, as opposed to the line of the operation start.
+        As CPython outputs the last bit of bytecode executed, and not the
+        line of the operation. For example calls, output the line of the
+        last argument, as opposed to the line of the operation start.
 
-            For tests, we wants to be compatible. In improved more, we are
-            not being fully compatible, and just drop it altogether.
+        For tests, we wants to be compatible. In improved more, we are
+        not being fully compatible, and just drop it altogether.
         """
 
         # Getting the same source reference can be dealt with quickly, so do
         # this first.
-        if self.source_ref is not source_ref and \
-           Options.isFullCompat() and \
-           self.source_ref != source_ref:
+        if (
+            self.source_ref is not source_ref
+            and Options.is_fullcompat
+            and self.source_ref != source_ref
+        ):
             # An attribute outside of "__init__", so we save one memory for the
             # most cases. Very few cases involve splitting across lines.
-            # pylint: disable=W0201
+            # false alarm for non-slot:
+            # pylint: disable=I0021,assigning-non-slot,attribute-defined-outside-init
             self.effective_source_ref = source_ref
 
-
     def getCompatibleSourceReference(self):
-        """ Bug compatible line numbers information.
+        """Bug compatible line numbers information.
 
-            See above.
+        See above.
         """
         return getattr(self, "effective_source_ref", self.source_ref)
 
     def asXml(self):
-        line = self.getSourceReference().getLineNumber()
+        line = self.source_ref.getLineNumber()
 
-        result = TreeXML.Element(
-            "node",
-            kind = self.__class__.__name__,
-            line = "%s" % line
-        )
+        result = TreeXML.Element("node", kind=self.__class__.__name__, line="%s" % line)
 
         compat_line = self.getCompatibleSourceReference().getLineNumber()
 
@@ -304,40 +285,33 @@ class NodeBase(NodeMetaClassBase):
             result.set(key, str(value))
 
         for name, children in self.getVisitableNodesNamed():
-            role = TreeXML.Element(
-                "role",
-                name = name
-            )
+            role = TreeXML.Element("role", name=name)
 
             result.append(role)
 
             if children is None:
                 role.attrib["type"] = "none"
             elif type(children) not in (list, tuple):
-                role.append(
-                    children.asXml()
-                )
+                role.append(children.asXml())
             else:
                 role.attrib["type"] = "list"
 
                 for child in children:
-                    role.append(
-                        child.asXml()
-                    )
+                    role.append(child.asXml())
 
         return result
 
     @classmethod
     def fromXML(cls, provider, source_ref, **args):
         # Only some things need a provider, pylint: disable=unused-argument
-        return cls(source_ref = source_ref, **args)
+        return cls(source_ref=source_ref, **args)
 
     def asXmlText(self):
         xml = self.asXml()
 
         return TreeXML.toString(xml)
 
-    def dump(self, level = 0):
+    def dump(self, level=0):
         Tracing.printIndented(level, self)
         Tracing.printSeparator(level)
 
@@ -368,35 +342,44 @@ class NodeBase(NodeMetaClassBase):
     def isExpressionConstantRef():
         return False
 
+    @staticmethod
+    def isExpressionOperationUnary():
+        return False
+
     @staticmethod
     def isExpressionOperationBinary():
         return False
 
-    def isExpressionSideEffects(self):
-        # Virtual method, pylint: disable=no-self-use
-
-        # We need to provide this, as these node kinds are only imported if
-        # necessary, but we test against them.
+    @staticmethod
+    def isExpressionOperationInplace():
         return False
 
-    def isExpressionMakeSequence(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def isExpressionComparison():
         return False
 
-    def isNumberConstant(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def isExpressionSideEffects():
         return False
 
-    def isExpressionCall(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def isExpressionMakeSequence():
         return False
 
-    def isExpressionFunctionBodyBase(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def isNumberConstant():
         return False
 
-    def isExpressionOutlineFunctionBodyBase(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def isExpressionCall():
+        return False
+
+    @staticmethod
+    def isExpressionFunctionBodyBase():
+        return False
+
+    @staticmethod
+    def isExpressionOutlineFunctionBase():
         return False
 
     def visit(self, context, visitor):
@@ -405,26 +388,29 @@ class NodeBase(NodeMetaClassBase):
         for visitable in self.getVisitableNodes():
             visitable.visit(context, visitor)
 
-    def getVisitableNodes(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def getVisitableNodes():
+
         return ()
 
-    def getVisitableNodesNamed(self):
-        """ Named children dictionary.
+    @staticmethod
+    def getVisitableNodesNamed():
+        """Named children dictionary.
 
-            For use in debugging and XML output.
+        For use in debugging and XML output.
         """
 
-        # Virtual method, pylint: disable=no-self-use
         return ()
 
-    def getName(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def getName():
+        """ Name of the node if any. """
+
         return None
 
-    def mayHaveSideEffects(self):
+    @staticmethod
+    def mayHaveSideEffects():
         """ Unless we are told otherwise, everything may have a side effect. """
-        # Virtual method, pylint: disable=no-self-use
 
         return True
 
@@ -436,21 +422,24 @@ class NodeBase(NodeMetaClassBase):
 
         return (self,)
 
-    def mayRaiseException(self, exception_type):
+    @staticmethod
+    def mayRaiseException(exception_type):
         """ Unless we are told otherwise, everything may raise everything. """
-        # Virtual method, pylint: disable=no-self-use,unused-argument
+        # Virtual method, pylint: disable=unused-argument
 
         return True
 
-    def mayReturn(self):
-        return "_RETURN" in self.kind
-
-    def mayBreak(self):
-        # For overload, pylint: disable=no-self-use
+    @staticmethod
+    def mayReturn():
+        """ May this node do a return exit, to be overloaded for things that might. """
         return False
 
-    def mayContinue(self):
-        # For overload, pylint: disable=no-self-use
+    @staticmethod
+    def mayBreak():
+        return False
+
+    @staticmethod
+    def mayContinue():
         return False
 
     def needsFrame(self):
@@ -458,20 +447,22 @@ class NodeBase(NodeMetaClassBase):
 
         return self.mayRaiseException(BaseException)
 
-    def willRaiseException(self, exception_type):
+    @staticmethod
+    def willRaiseException(exception_type):
         """ Unless we are told otherwise, nothing may raise anything. """
-        # Virtual method, pylint: disable=no-self-use,unused-argument
-
+        # Virtual method, pylint: disable=unused-argument
         return False
 
-    def isStatementAborting(self):
+    @staticmethod
+    def isStatementAborting():
         """ Is the node aborting, control flow doesn't continue after this node.  """
-        assert self.isStatement(), self.kind
-
         return False
 
 
 class CodeNodeMixin(object):
+    # Mixins are not allow to specify slots, pylint: disable=assigning-non-slot
+    __slots__ = ()
+
     def __init__(self, name, code_prefix):
         assert name is not None
 
@@ -497,15 +488,14 @@ class CodeNodeMixin(object):
             assert isinstance(self, CodeNodeMixin)
 
             if self.name:
-                name = uid + '_' + self.name.strip("<>")
+                name = uid + "_" + self.name.strip("<>")
             else:
                 name = uid
 
-            self.code_name = "%s$$$%s%s" % (
-                parent_name,
-                self.code_prefix,
-                name
-            )
+            if str is not bytes:
+                name = name.encode("ascii", "c_identifier").decode()
+
+            self.code_name = "%s$$$%s_%s" % (parent_name, self.code_prefix, name)
 
         return self.code_name
 
@@ -519,45 +509,51 @@ class CodeNodeMixin(object):
 
 
 class ChildrenHavingMixin(object):
+    # Mixins are not allow to specify slots.
+    __slots__ = ()
+
     named_children = ()
 
     checkers = {}
 
     def __init__(self, values):
-        assert type(self.named_children) is tuple and self.named_children
+        assert (
+            type(self.named_children) is tuple and self.named_children
+        ), self.named_children
 
         # TODO: Make this true.
         # assert len(self.named_children) > 1, self.kind
 
         # Check for completeness of given values, everything should be there
         # but of course, might be put to None.
-        assert set(values.keys()) == set(self.named_children)
+        if set(values.keys()) != set(self.named_children):
+            raise NuitkaNodeDesignError(
+                "Must pass named children in value dictionary",
+                set(values.keys()),
+                set(self.named_children),
+            )
 
         for name, value in values.items():
             if name in self.checkers:
                 value = self.checkers[name](value)
 
-            assert type(value) is not list, name
-
             if type(value) is tuple:
                 assert None not in value, name
 
                 for val in value:
                     val.parent = self
-            elif value is not None:
-                value.parent = self
             elif value is None:
                 pass
             else:
-                assert False, type(value)
+                value.parent = self
 
             attr_name = "subnode_" + name
             setattr(self, attr_name, value)
 
     def setChild(self, name, value):
-        """ Set a child value.
+        """Set a child value.
 
-            Do not overload, provider self.checkers instead.
+        Do not overload, provider self.checkers instead.
         """
         # Only accept legal child names
         assert name in self.named_children, name
@@ -578,33 +574,33 @@ class ChildrenHavingMixin(object):
 
         attr_name = "subnode_" + name
 
-        # Determine old value, and inform it about loosing its parent.
+        # Determine old value, and inform it about losing its parent.
         old_value = getattr(self, attr_name)
 
         assert old_value is not value, value
 
         setattr(self, attr_name, value)
 
+    def clearChild(self, name):
+        # Only accept legal child names
+        assert name in self.named_children, name
+
+        if name in self.checkers:
+            self.checkers[name](None)
+
+        attr_name = "subnode_" + name
+
+        # Determine old value, and inform it about losing its parent.
+        old_value = getattr(self, attr_name)
+
+        assert old_value is not None
+
+        setattr(self, attr_name, None)
+
     def getChild(self, name):
         attr_name = "subnode_" + name
         return getattr(self, attr_name)
 
-    @staticmethod
-    def childGetter(name):
-        attr_name = intern("subnode_" + name)
-
-        def getter(self):
-            return getattr(self, attr_name)
-
-        return getter
-
-    @staticmethod
-    def childSetter(name):
-        def setter(self, value):
-            self.setChild(name, value)
-
-        return setter
-
     def getVisitableNodes(self):
         # TODO: Consider if a generator would be faster.
         result = []
@@ -622,16 +618,15 @@ class ChildrenHavingMixin(object):
                 result.append(value)
             else:
                 raise AssertionError(
-                    self,
-                    "has illegal child", name, value, value.__class__
+                    self, "has illegal child", name, value, value.__class__
                 )
 
         return tuple(result)
 
     def getVisitableNodesNamed(self):
-        """ Named children dictionary.
+        """Named children dictionary.
 
-            For use in debugging and XML output.
+        For use in debugging and XML output.
         """
         for name in self.named_children:
             attr_name = "subnode_" + name
@@ -639,7 +634,6 @@ class ChildrenHavingMixin(object):
 
             yield name, value
 
-
     def replaceChild(self, old_node, new_node):
         if new_node is not None and not isinstance(new_node, NodeBase):
             raise AssertionError(
@@ -661,19 +655,12 @@ class ChildrenHavingMixin(object):
                             key,
                             tuple(
                                 (val if val is not old_node else new_node)
-                                for val in
-                                value
-                            )
+                                for val in value
+                            ),
                         )
                     else:
                         self.setChild(
-                            key,
-                            tuple(
-                                val
-                                for val in
-                                value
-                                if val is not old_node
-                            )
+                            key, tuple(val for val in value if val is not old_node)
                         )
 
                     return key
@@ -685,12 +672,7 @@ class ChildrenHavingMixin(object):
             else:
                 assert False, (key, value, value.__class__)
 
-        raise AssertionError(
-            "Didn't find child",
-            old_node,
-            "in",
-            self
-        )
+        raise AssertionError("Didn't find child", old_node, "in", self)
 
     def getCloneArgs(self):
         values = {}
@@ -703,17 +685,11 @@ class ChildrenHavingMixin(object):
             if value is None:
                 values[key] = None
             elif type(value) is tuple:
-                values[key] = tuple(
-                    v.makeClone()
-                    for v in
-                    value
-                )
+                values[key] = tuple(v.makeClone() for v in value)
             else:
                 values[key] = value.makeClone()
 
-        values.update(
-            self.getDetails()
-        )
+        values.update(self.getDetails())
 
         return values
 
@@ -725,16 +701,13 @@ class ChildrenHavingMixin(object):
 
 
 class ClosureGiverNodeMixin(CodeNodeMixin):
-    """ Blass class for nodes that provide variables for closure takers. """
-    def __init__(self, name, code_prefix):
-        CodeNodeMixin.__init__(
-            self,
-            name        = name,
-            code_prefix = code_prefix
-        )
+    """Base class for nodes that provide variables for closure takers. """
 
-        self.providing = {}
-        self.variable_order = []
+    # Mixins are not allow to specify slots, pylint: disable=assigning-non-slot
+    __slots__ = ()
+
+    def __init__(self, name, code_prefix):
+        CodeNodeMixin.__init__(self, name=name, code_prefix=code_prefix)
 
         self.temp_variables = {}
 
@@ -743,47 +716,27 @@ class ClosureGiverNodeMixin(CodeNodeMixin):
         self.preserver_id = 0
 
     def hasProvidedVariable(self, variable_name):
-        return variable_name in self.providing
+        return self.locals_scope.hasProvidedVariable(variable_name)
 
     def getProvidedVariable(self, variable_name):
-        if variable_name not in self.providing:
-            self.providing[variable_name] = self.createProvidedVariable(
-                variable_name = variable_name
-            )
-            self.variable_order.append(variable_name)
+        if not self.locals_scope.hasProvidedVariable(variable_name):
+            variable = self.createProvidedVariable(variable_name=variable_name)
+            self.locals_scope.registerProvidedVariable(variable)
 
-        return self.providing[variable_name]
+        return self.locals_scope.getProvidedVariable(variable_name)
 
+    @abstractmethod
     def createProvidedVariable(self, variable_name):
-        # Virtual method, pylint: disable=no-self-use
-        assert type(variable_name) is str
-
-        return None
-
-    def registerProvidedVariable(self, variable):
-        assert variable is not None
-
-        variable_name = variable.getName()
-        self.providing[variable_name] = variable
-        self.variable_order.append(variable_name)
-
-    def getProvidedVariableOrder(self):
-        return self.variable_order
+        """Create a variable provided by this function."""
 
     def allocateTempScope(self, name):
         self.temp_scopes[name] = self.temp_scopes.get(name, 0) + 1
 
-        return "%s_%d" % (
-            name,
-            self.temp_scopes[name]
-        )
+        return "%s_%d" % (name, self.temp_scopes[name])
 
-    def allocateTempVariable(self, temp_scope, name):
+    def allocateTempVariable(self, temp_scope, name, temp_type=None):
         if temp_scope is not None:
-            full_name = "%s__%s" % (
-                temp_scope,
-                name
-            )
+            full_name = "%s__%s" % (temp_scope, name)
         else:
             assert name != "result"
 
@@ -792,9 +745,7 @@ class ClosureGiverNodeMixin(CodeNodeMixin):
         # No duplicates please.
         assert full_name not in self.temp_variables, full_name
 
-        result = self.createTempVariable(
-            temp_name = full_name
-        )
+        result = self.createTempVariable(temp_name=full_name, temp_type=temp_type)
 
         # Late added temp variables should be treated with care for the
         # remaining trace.
@@ -803,14 +754,18 @@ class ClosureGiverNodeMixin(CodeNodeMixin):
 
         return result
 
-    def createTempVariable(self, temp_name):
+    def createTempVariable(self, temp_name, temp_type):
         if temp_name in self.temp_variables:
-            return self.temp_variables[ temp_name ]
+            return self.temp_variables[temp_name]
 
-        result = Variables.TempVariable(
-            owner         = self,
-            variable_name = temp_name
-        )
+        if temp_type is None:
+            temp_class = Variables.TempVariable
+        elif temp_type == "bool":
+            temp_class = Variables.TempVariableBool
+        else:
+            assert False, temp_type
+
+        result = temp_class(owner=self, variable_name=temp_name)
 
         self.temp_variables[temp_name] = result
 
@@ -827,11 +782,23 @@ class ClosureGiverNodeMixin(CodeNodeMixin):
     def getTempVariables(self):
         return self.temp_variables.values()
 
-    def removeTempVariable(self, variable):
+    def _removeTempVariable(self, variable):
         del self.temp_variables[variable.getName()]
 
+    def optimizeUnusedTempVariables(self):
+        remove = []
+
+        for temp_variable in self.getTempVariables():
+            empty = self.trace_collection.hasEmptyTraces(variable=temp_variable)
+
+            if empty:
+                remove.append(temp_variable)
+
+        for temp_variable in remove:
+            self._removeTempVariable(temp_variable)
+
     def allocatePreserverId(self):
-        if python_version >= 300:
+        if python_version >= 0x300:
             self.preserver_id += 1
 
         return self.preserver_id
@@ -840,6 +807,9 @@ class ClosureGiverNodeMixin(CodeNodeMixin):
 class ClosureTakerMixin(object):
     """ Mixin for nodes that accept variables from closure givers. """
 
+    # Mixins are not allow to specify slots, pylint: disable=assigning-non-slot
+    __slots__ = ()
+
     def __init__(self, provider):
         self.provider = provider
 
@@ -849,9 +819,7 @@ class ClosureTakerMixin(object):
         return self.provider
 
     def getClosureVariable(self, variable_name):
-        result = self.provider.getVariableForClosure(
-            variable_name = variable_name
-        )
+        result = self.provider.getVariableForClosure(variable_name=variable_name)
         assert result is not None, variable_name
 
         if not result.isModuleVariable():
@@ -867,13 +835,8 @@ class ClosureTakerMixin(object):
     def getClosureVariables(self):
         return tuple(
             sorted(
-                [
-                    take
-                    for take in
-                    self.taken
-                    if not take.isModuleVariable()
-                ],
-                key = lambda x : x.getName()
+                [take for take in self.taken if not take.isModuleVariable()],
+                key=lambda x: x.getName(),
             )
         )
 
@@ -902,36 +865,33 @@ class ClosureTakerMixin(object):
 
 
 class StatementBase(NodeBase):
-    """ Base class for all statements.
+    """Base class for all statements."""
 
-    """
     # Base classes can be abstract, pylint: disable=abstract-method
 
     # TODO: Have them all.
     # @abstractmethod
-    def getStatementNiceName(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def getStatementNiceName():
         return "undescribed statement"
 
     def computeStatementSubExpressions(self, trace_collection):
-        """ Compute a statement.
+        """Compute a statement.
 
-            Default behavior is to just visit the child expressions first, and
-            then the node "computeStatement". For a few cases this needs to
-            be overloaded.
+        Default behavior is to just visit the child expressions first, and
+        then the node "computeStatement". For a few cases this needs to
+        be overloaded.
         """
         expressions = self.getVisitableNodes()
 
         for count, expression in enumerate(expressions):
             assert expression.isExpression(), (self, expression)
 
-            expression = trace_collection.onExpression(
-                expression = expression
-            )
+            expression = trace_collection.onExpression(expression=expression)
 
             if expression.willRaiseException(BaseException):
                 wrapped_expression = makeStatementOnlyNodesFromExpressions(
-                    expressions[:count+1]
+                    expressions[: count + 1]
                 )
 
                 assert wrapped_expression is not None
@@ -939,10 +899,8 @@ class StatementBase(NodeBase):
                 return (
                     wrapped_expression,
                     "new_raise",
-                    lambda : "For %s the child expression '%s' will raise." % (
-                        self.getStatementNiceName(),
-                        expression.getChildNameNice()
-                    )
+                    lambda: "For %s the child expression '%s' will raise."
+                    % (self.getStatementNiceName(), expression.getChildNameNice()),
                 )
 
         return self, None, None
@@ -950,12 +908,9 @@ class StatementBase(NodeBase):
 
 class StatementChildrenHavingBase(ChildrenHavingMixin, StatementBase):
     def __init__(self, values, source_ref):
-        StatementBase.__init__(self, source_ref = source_ref)
+        StatementBase.__init__(self, source_ref=source_ref)
 
-        ChildrenHavingMixin.__init__(
-            self,
-            values = values
-        )
+        ChildrenHavingMixin.__init__(self, values=values)
 
 
 class StatementChildHavingBase(StatementBase):
@@ -964,7 +919,7 @@ class StatementChildHavingBase(StatementBase):
     checker = None
 
     def __init__(self, value, source_ref):
-        StatementBase.__init__(self, source_ref = source_ref)
+        StatementBase.__init__(self, source_ref=source_ref)
 
         assert type(self.named_child) is str and self.named_child
 
@@ -989,9 +944,9 @@ class StatementChildHavingBase(StatementBase):
         setattr(self, attr_name, value)
 
     def setChild(self, name, value):
-        """ Set a child value.
+        """Set a child value.
 
-            Do not overload, provider self.checkers instead.
+        Do not overload, provider self.checkers instead.
         """
         # Only accept legal child names
         assert name == self.named_child, name
@@ -1012,7 +967,7 @@ class StatementChildHavingBase(StatementBase):
 
         attr_name = "subnode_" + name
 
-        # Determine old value, and inform it about loosing its parent.
+        # Determine old value, and inform it about losing its parent.
         old_value = getattr(self, attr_name)
 
         assert old_value is not value, value
@@ -1024,22 +979,6 @@ class StatementChildHavingBase(StatementBase):
         attr_name = "subnode_" + name
         return getattr(self, attr_name)
 
-    @staticmethod
-    def childGetter(name):
-        attr_name = "subnode_" + name
-
-        def getter(self):
-            return getattr(self, attr_name)
-
-        return getter
-
-    @staticmethod
-    def childSetter(name):
-        def setter(self, value):
-            self.setChild(name, value)
-
-        return setter
-
     def getVisitableNodes(self):
         # TODO: Consider if a generator would be faster.
         attr_name = "subnode_" + self.named_child
@@ -1052,15 +991,12 @@ class StatementChildHavingBase(StatementBase):
         elif isinstance(value, NodeBase):
             return (value,)
         else:
-            raise AssertionError(
-                self,
-                "has illegal child", value, value.__class__
-            )
+            raise AssertionError(self, "has illegal child", value, value.__class__)
 
     def getVisitableNodesNamed(self):
-        """ Named children dictionary.
+        """Named children dictionary.
 
-            For use in debugging and XML output.
+        For use in debugging and XML output.
         """
         attr_name = "subnode_" + self.named_child
         value = getattr(self, attr_name)
@@ -1087,20 +1023,12 @@ class StatementChildHavingBase(StatementBase):
                     self.setChild(
                         key,
                         tuple(
-                            (val if val is not old_node else new_node)
-                            for val in
-                            value
-                        )
+                            (val if val is not old_node else new_node) for val in value
+                        ),
                     )
                 else:
                     self.setChild(
-                        key,
-                        tuple(
-                            val
-                            for val in
-                            value
-                            if val is not old_node
-                        )
+                        key, tuple(val for val in value if val is not old_node)
                     )
 
                 return key
@@ -1112,12 +1040,7 @@ class StatementChildHavingBase(StatementBase):
         else:
             assert False, (key, value, value.__class__)
 
-        raise AssertionError(
-            "Didn't find child",
-            old_node,
-            "in",
-            self
-        )
+        raise AssertionError("Didn't find child", old_node, "in", self)
 
     def getCloneArgs(self):
         # Make clones of child nodes too.
@@ -1131,28 +1054,28 @@ class StatementChildHavingBase(StatementBase):
         if value is None:
             values[key] = None
         elif type(value) is tuple:
-            values[key] = tuple(
-                v.makeClone()
-                for v in
-                value
-            )
+            values[key] = tuple(v.makeClone() for v in value)
         else:
             values[key] = value.makeClone()
 
-        values.update(
-            self.getDetails()
-        )
+        values.update(self.getDetails())
 
         return values
 
     def finalize(self):
         del self.parent
 
-        for c in self.getVisitableNodes():
-            c.finalize()
+        attr_name = "subnode_" + self.named_child
+        child = getattr(self, attr_name)
+        if child is not None:
+            child.finalize()
+        delattr(self, attr_name)
 
 
 class SideEffectsFromChildrenMixin(object):
+    # Mixins are not allow to specify slots.
+    __slots__ = ()
+
     def mayHaveSideEffects(self):
         for child in self.getVisitableNodes():
             if child.mayHaveSideEffects():
@@ -1165,9 +1088,7 @@ class SideEffectsFromChildrenMixin(object):
         result = []
 
         for child in self.getVisitableNodes():
-            result.extend(
-                child.extractSideEffects()
-            )
+            result.extend(child.extractSideEffects())
 
         return tuple(result)
 
@@ -1182,13 +1103,13 @@ class SideEffectsFromChildrenMixin(object):
             return (
                 makeStatementOnlyNodesFromExpressions(side_effects),
                 "new_statements",
-                "Reduced unused %s to side effects." % self.kind
+                "Lowered unused expression %s to its side effects." % self.kind,
             )
         else:
             return (
                 None,
                 "new_statements",
-                "Removed %s without side effects." % self.kind
+                "Removed %s without side effects." % self.kind,
             )
 
 
@@ -1197,22 +1118,13 @@ def makeChild(provider, child, source_ref):
 
     if child_type == "list":
         return [
-            fromXML(
-                provider   = provider,
-                xml        = sub_child,
-                source_ref = source_ref
-            )
-            for sub_child in
-            child
+            fromXML(provider=provider, xml=sub_child, source_ref=source_ref)
+            for sub_child in child
         ]
     elif child_type == "none":
         return None
     else:
-        return fromXML(
-            provider   = provider,
-            xml        = child[0],
-            source_ref = source_ref
-        )
+        return fromXML(provider=provider, xml=child[0], source_ref=source_ref)
 
 
 def getNodeClassFromKind(kind):
@@ -1227,8 +1139,7 @@ def extractKindAndArgsFromXML(xml, source_ref):
 
     if source_ref is None:
         source_ref = SourceCodeReference.fromFilenameAndLine(
-            args["filename"],
-            int(args["line"])
+            args["filename"], int(args["line"])
         )
 
         del args["filename"]
@@ -1243,7 +1154,7 @@ def extractKindAndArgsFromXML(xml, source_ref):
     return kind, node_class, args, source_ref
 
 
-def fromXML(provider, xml, source_ref = None):
+def fromXML(provider, xml, source_ref=None):
     assert xml.tag == "node", xml
 
     kind, node_class, args, source_ref = extractKindAndArgsFromXML(xml, source_ref)
@@ -1253,9 +1164,13 @@ def fromXML(provider, xml, source_ref = None):
         # global stream     instead. For now, this will do. pylint: disable=eval-used
         args["constant"] = eval(args["constant"])
 
-    if kind in ("ExpressionFunctionBody", "PythonMainModule",
-                "PythonCompiledModule", "PythonCompiledPackage",
-                "PythonInternalModule"):
+    if kind in (
+        "ExpressionFunctionBody",
+        "PythonMainModule",
+        "PythonCompiledModule",
+        "PythonCompiledPackage",
+        "PythonInternalModule",
+    ):
         delayed = node_class.named_children
 
         if "code_flags" in args:
@@ -1277,11 +1192,7 @@ def fromXML(provider, xml, source_ref = None):
             args[child_name] = child
 
     try:
-        return node_class.fromXML(
-            provider   = provider,
-            source_ref = source_ref,
-            **args
-        )
+        return node_class.fromXML(provider=provider, source_ref=source_ref, **args)
     except (TypeError, AttributeError):
         Tracing.printLine(node_class, args, source_ref)
         raise
