@@ -86,6 +86,7 @@ not exist, a list of all available will be given.""",
                         "mediaservice",
                         "printsupport",
                         "platforms",
+                        "styles",
                     )
                     if self.hasPluginFamily(family)
                 )
@@ -112,7 +113,7 @@ import %(binding_name)s.QtCore
         )
 
         return self.queryRuntimeInformationMultiple(
-            applyBindingName("%(binding_name)s_info"),
+            info_name=applyBindingName("%(binding_name)s_info"),
             setup_codes=setup_codes,
             values=(
                 (
@@ -141,6 +142,12 @@ import %(binding_name)s.QtCore
                         else "%(binding_name)s.QtCore.PYQT_VERSION_STR"
                     ),
                 ),
+                (
+                    "nuitka_patch_level",
+                    applyBindingName(
+                        "getattr(%(binding_name)s, '_nuitka_patch_level', 0)"
+                    ),
+                ),
             ),
         )
 
@@ -148,6 +155,10 @@ import %(binding_name)s.QtCore
         """ Get the version of the binding in tuple digit form, e.g. (6,0,3) """
         return self._getQtInformation().version
 
+    def _getNuitkaPatchLevel(self):
+        """ Does it include the Nuitka patch, i.e. is a self-built one with it applied. """
+        return self._getQtInformation().nuitka_patch_level
+
     def getQtPluginDirs(self):
         if self.qt_plugins_dirs is not None:
             return self.qt_plugins_dirs
@@ -432,8 +443,7 @@ Setting Qt library path to distribution folder. We need to avoid loading target
 system Qt plug-ins, which may be from another Qt version.""",
             )
 
-    @staticmethod
-    def createPreModuleLoadCode(module):
+    def createPreModuleLoadCode(self, module):
         """Method called when a module is being imported.
 
         Notes:
@@ -445,23 +455,23 @@ system Qt plug-ins, which may be from another Qt version.""",
         Returns:
             Code to insert and descriptive text (tuple), or (None, None).
         """
-        if (
-            not isWin32Windows() or not Options.isStandaloneMode()
-        ):  # we are only relevant on standalone mode for Windows
+
+        # This isonly relevant on standalone mode for Windows
+        if not isWin32Windows() or not Options.isStandaloneMode():
             return None
 
-        if module.getFullName() != "PyQt5":
-            return None
+        full_name = module.getFullName()
 
-        code = """import os
+        if full_name == self.binding_name:
+            code = """import os
 path = os.environ.get("PATH", "")
 if not path.startswith(__nuitka_binary_dir):
     os.environ["PATH"] = __nuitka_binary_dir + ";" + path
 """
-        return (
-            code,
-            "Adding binary folder to runtime 'PATH' environment variable for proper loading.",
-        )
+            return (
+                code,
+                "Adding binary folder to runtime 'PATH' environment variable for proper loading.",
+            )
 
     def considerExtraDlls(self, dist_dir, module):
         # pylint: disable=too-many-branches,too-many-locals,too-many-statements
@@ -691,8 +701,16 @@ class NuitkaPluginPySide2Plugins(NuitkaPluginQtBindingsPluginBase):
     binding_name = "PySide2"
 
     def __init__(self, qt_plugins):
-        if python_version < 0x360:
-            self.sysexit("Error, PySide2 is not supported with Nuitka on CPython <3.6.")
+        if self._getNuitkaPatchLevel() < 1:
+            self.warning(
+                """\
+This PySide2 version only partially supported through workarounds, full support: https://nuitka.net/pages/pyside2.html"""
+            )
+
+            if python_version < 0x360:
+                self.sysexit(
+                    "Error, unpatched PySide2 is not supported before CPython <3.6."
+                )
 
         NuitkaPluginQtBindingsPluginBase.__init__(self, qt_plugins)
 
@@ -717,7 +735,10 @@ class NuitkaPluginPySide2Plugins(NuitkaPluginQtBindingsPluginBase):
         if result:
             return result
 
-        if module.getFullName() == self.binding_name:
+        if (
+            self._getNuitkaPatchLevel() < 1
+            and module.getFullName() == self.binding_name
+        ):
             code = r"""\
 # Make them unique and count them.
 wrapper_count = 0
@@ -796,7 +817,7 @@ PySide2.QtCore.QObject.__init_subclass__ = my_init_subclass
             return (
                 code,
                 """\
-Money patching classes derived from PySide2 base classes to pass PySide2 checks.""",
+Monkey patching classes derived from PySide2 base classes to pass PySide2 checks.""",
             )
 
 
@@ -823,9 +844,10 @@ class NuitkaPluginPySide6Plugins(NuitkaPluginQtBindingsPluginBase):
     def __init__(self, qt_plugins):
         NuitkaPluginQtBindingsPluginBase.__init__(self, qt_plugins)
 
-        if self._getBindingVersion() < (6, 0, 3):
+        if self._getBindingVersion() < (6, 1):
             self.warning(
-                "The version of PySide6 you are using should be 6.0.3 or higher, otherwise callbacks won't work."
+                """\
+Only PySide 6.1 or higher (or dev branch compiled), otherwise callbacks won't work."""
             )
 
     def _getQmlTargetDir(self, target_plugin_dir):
